@@ -3,11 +3,11 @@
 # referenced from other places in the experiment
 from functools import wraps
 import inspect
-from typing import List, Union
+from typing import Any, List, Dict, Union
 
 from logzero import logger
 
-from chaoslib.types import Activity, Experiment, Settings
+from chaoslib.types import Activity, Experiment, Settings, Schedule, Strategy
 
 
 __all__ = ["cache_activities", "clear_cache", "lookup_activity", "with_cache"]
@@ -49,7 +49,11 @@ def with_cache(f):
     function.
     """
     @wraps(f)
-    def wrapped(experiment: Experiment, settings: Settings = None):
+    def wrapped(experiment: Experiment, settings: Settings = None,
+                experiment_vars: Dict[str, Any] = None,
+                strategy: Strategy = Strategy.DEFAULT,
+                schedule: Schedule = None,
+                event_handlers: List['RunEventHandler'] = None):
         try:
             if experiment:
                 cache_activities(experiment)
@@ -58,8 +62,17 @@ def with_cache(f):
             arguments = {
                 "experiment": experiment
             }
+
             if "settings" in sig.parameters:
                 arguments["settings"] = settings
+            if "experiment_vars" in sig.parameters:
+                arguments["experiment_vars"] = experiment_vars
+            if "strategy" in sig.parameters:
+                arguments["strategy"] = strategy
+            if "schedule" in sig.parameters:
+                arguments["schedule"] = schedule
+            if "event_handlers" in sig.parameters:
+                arguments["event_handlers"] = event_handlers
             return f(**arguments)
         finally:
             clear_cache()
