@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-from concurrent.futures import Future, ThreadPoolExecutor
+from concurrent.futures import Future, ThreadPoolExecutor, TimeoutError
 from datetime import datetime
 import platform
 import threading
@@ -13,6 +13,7 @@ from chaoslib.activity import run_activities
 from chaoslib.control import initialize_controls, controls, cleanup_controls, \
     Control, initialize_global_controls, cleanup_global_controls
 from chaoslib.exceptions import ChaosException, InterruptExecution
+from chaoslib.exit import exit_signals
 from chaoslib.configuration import load_configuration
 from chaoslib.hypothesis import run_steady_state_hypothesis
 from chaoslib.rollback import run_rollbacks
@@ -296,10 +297,11 @@ class Runner:
             journal: Journal = None) -> Journal:
 
         self.configure(experiment, settings)
-        journal = self._run(
-            self.strategy, self.schedule, experiment, journal,
-            self.config, self.secrets, self.settings, experiment_vars,
-            self.event_registry)
+        with exit_signals():
+            journal = self._run(
+                self.strategy, self.schedule, experiment, journal,
+                self.config, self.secrets, self.settings, experiment_vars,
+                self.event_registry)
         return journal
 
     def _run(self, strategy: Strategy, schedule: Schedule,  # noqa: C901
@@ -334,6 +336,13 @@ class Runner:
             "rollbacks", {}).get("strategy", "default")
         logger.info("Rollbacks strategy: {}".format(rollback_strategy))
 
+        exit_gracefully_with_rollbacks = True
+        with_ssh = has_steady_state_hypothesis_with_probes(experiment)
+        if not with_ssh:
+            logger.info(
+                "No steady state hypothesis defined. That's ok, just "
+                "exploring.")
+
         try:
             try:
                 control.begin(
@@ -341,25 +350,25 @@ class Runner:
                     secrets)
 
                 state = object()
-                if should_run_before_method(strategy):
+                if with_ssh and should_run_before_method(strategy):
                     state = run_gate_hypothesis(
                         experiment, journal, configuration, secrets,
                         event_registry, dry)
 
                 if state is not None:
-                    if should_run_during_method(strategy):
+                    if with_ssh and should_run_during_method(strategy):
                         run_hypothesis_during_method(
                             hypo_pool, continous_hypo_event, strategy,
                             schedule, experiment, journal, configuration,
                             secrets, event_registry, dry)
 
                     state = run_method(
-                         strategy, activity_pool, experiment, journal,
-                         configuration, secrets, event_registry, dry)
+                            strategy, activity_pool, experiment, journal,
+                            configuration, secrets, event_registry, dry)
 
                     continous_hypo_event.set()
                     if journal["status"] not in ["interrupted", "aborted"]:
-                        if state is not None and \
+                        if with_ssh and (state is not None) and \
                                 should_run_after_method(strategy):
                             run_deviation_validation_hypothesis(
                                 experiment, journal, configuration, secrets,
@@ -368,17 +377,27 @@ class Runner:
                 journal["status"] = "interrupted"
                 logger.fatal(str(i))
                 event_registry.interrupted(experiment, journal)
-            except (KeyboardInterrupt, SystemExit):
+            except KeyboardInterrupt:
                 journal["status"] = "interrupted"
-                logger.warning("Received an exit signal, "
-                               "leaving without applying rollbacks.")
+                logger.warning("Received a termination signal (Ctrl-C)...")
+                event_registry.signal_exit()
+            except SystemExit as x:
+                journal["status"] = "interrupted"
+                logger.warning("Received the exit signal: {}".format(x.code))
+
+                exit_gracefully_with_rollbacks = x.code != 30
+                if not exit_gracefully_with_rollbacks:
+                    logger.warning("Ignoring rollbacks as per signal")
                 event_registry.signal_exit()
             finally:
                 hypo_pool.shutdown(wait=True)
 
-            run_rollback(
-                rollback_strategy, rollback_pool, experiment, journal,
-                configuration, secrets, event_registry, dry)
+            # just in case a signal overrode everything else to tell us not to
+            # play them anyway (see the exit.py module)
+            if exit_gracefully_with_rollbacks:
+                run_rollback(
+                    rollback_strategy, rollback_pool, experiment, journal,
+                    configuration, secrets, event_registry, dry)
 
             journal["end"] = datetime.utcnow().isoformat()
             journal["duration"] = time.time() - started_at
@@ -441,6 +460,7 @@ def run_gate_hypothesis(experiment: Experiment, journal: Journal,
     state = run_steady_state_hypothesis(
         experiment, configuration, secrets, dry=dry)
     journal["steady_states"]["before"] = state
+    event_registry.hypothesis_before_completed(experiment, state, journal)
     if state is not None and not state["steady_state_met"]:
         journal["steady_states"]["before"] = state
         journal["status"] = "failed"
@@ -450,7 +470,7 @@ def run_gate_hypothesis(experiment: Experiment, journal: Journal,
             "Steady state probe '{p}' is not in the given "
             "tolerance so failing this experiment".format(
                 p=p["activity"]["name"]))
-    event_registry.hypothesis_before_completed(experiment, state, journal)
+        return
     return state
 
 
@@ -470,6 +490,7 @@ def run_deviation_validation_hypothesis(experiment: Experiment,
     state = run_steady_state_hypothesis(
         experiment, configuration, secrets, dry=dry)
     journal["steady_states"]["after"] = state
+    event_registry.hypothesis_after_completed(experiment, state, journal)
     if state is not None and \
             not state["steady_state_met"]:
         journal["deviated"] = True
@@ -480,7 +501,6 @@ def run_deviation_validation_hypothesis(experiment: Experiment,
             "given tolerance so failing this "
             "experiment".format(
                 p=p["activity"]["name"]))
-    event_registry.hypothesis_after_completed(experiment, state, journal)
     return state
 
 
@@ -522,13 +542,13 @@ def run_method(strategy: Strategy, activity_pool: ThreadPoolExecutor,
                configuration: Configuration, secrets: Secrets,
                event_registry: EventHandlerRegistry,
                dry: bool = False) -> Optional[List[Run]]:
+    logger.info("Playing your experiment's method now...")
     event_registry.start_method(experiment)
     try:
         state = apply_activities(
             experiment, configuration, secrets, activity_pool, journal, dry)
-        journal["run"] = state
         event_registry.method_completed(experiment, state)
-        return journal["run"]
+        return state
     except InterruptExecution:
         event_registry.method_completed(experiment)
         raise
@@ -710,27 +730,66 @@ def apply_activities(experiment: Experiment, configuration: Configuration,
                      journal: Journal, dry: bool = False) -> List[Run]:
     with controls(level="method", experiment=experiment, context=experiment,
                   configuration=configuration, secrets=secrets) as control:
-        runs = []
-        for run in run_activities(
-                experiment, configuration, secrets, pool, dry):
-            runs.append(run)
-            if journal["status"] in ["aborted", "failed", "interrupted"]:
-                break
-
-        if pool:
-            logger.debug("Waiting for background activities to complete...")
-            pool.shutdown(wait=True)
-
         result = []
-        for run in runs:
-            if not run:
-                continue
-            if isinstance(run, dict):
-                result.append(run)
-            else:
-                result.append(run.result())
+        runs = []
+        method = experiment.get("method", [])
+        wait_for_background_activities = True
 
-        control.with_state(result)
+        try:
+            for run in run_activities(
+                    experiment, configuration, secrets, pool, dry):
+                runs.append(run)
+                if journal["status"] in ["aborted", "failed", "interrupted"]:
+                    break
+        except SystemExit as x:
+            # when we got a signal for an ungraceful exit, we can decide
+            # not to wait for background activities. Their statuses will
+            # remain failed.
+            wait_for_background_activities = x.code != 30  # see exit.py
+            raise
+        finally:
+            background_activity_timeout = None
+
+            if wait_for_background_activities and pool:
+                logger.debug("Waiting for background activities to complete")
+                pool.shutdown(wait=True)
+            elif pool:
+                logger.debug(
+                    "Do not wait for the background activities to finish "
+                    "as per signal")
+                background_activity_timeout = 0.1
+                pool.shutdown(wait=False)
+
+            for index, run in enumerate(runs):
+                if not run:
+                    continue
+
+                if isinstance(run, dict):
+                    result.append(run)
+                else:
+                    try:
+                        # background activities
+                        result.append(
+                            run.result(timeout=background_activity_timeout))
+                    except TimeoutError:
+                        # we want an entry for the background activity in our
+                        # results anyway, we won't have anything meaningful
+                        # to say about it
+                        result.append({
+                            "activity": method[index],
+                            "status": "failed",
+                            "output": None,
+                            "duration": None,
+                            "start": None,
+                            "end": None,
+                            "exception": None
+                        })
+
+            # now let's ensure the journal has all activities in their correct
+            # order (background ones included)
+            journal["run"] = result
+
+            control.with_state(result)
 
     return result
 
@@ -760,3 +819,12 @@ def apply_rollbacks(experiment: Experiment, configuration: Configuration,
         control.with_state(result)
 
     return result
+
+
+def has_steady_state_hypothesis_with_probes(experiment: Experiment) -> bool:
+    steady_state_hypothesis = experiment.get("steady-state-hypothesis")
+    if steady_state_hypothesis:
+        probes = steady_state_hypothesis.get("probes")
+        if probes:
+            return len(probes) > 0
+    return False
